#include <stdio.h>
#include <stdlib.h>
#include "lrz.h"

// Node type constants
#define NODE_UNKNOWN   0
#define NODE_BRANCHING 1
#define NODE_VALUE     2

// Decoder state constants
#define STATE_RAW  0
#define STATE_COPY 1

// Linked list data structure for binary tree nodes
typedef struct LRZ_NODE LRZ_NODE;
struct  LRZ_NODE {
    int type, value, done;
    LRZ_NODE *parent, *child[2];
};

// Atomic data processing unit for file chunks
typedef struct {

    // Input/output properties
    unsigned char *src, *dst;
    int srclen, dstlen, srcoff, dstoff;

    // Binary trees
    LRZ_NODE *rawtree, *lentree, *disttree;

    // Bit processing properties
    int bitsleft, reg, subbits, state;
} LRZ_CHUNK;



////////////////////////////////////////////////////////////////////////////////
//                             Decoder Functions                              //
////////////////////////////////////////////////////////////////////////////////

// This macro is used in lrzdFreeTree() to handle deletion of individual nodes
#define DeleteNode() {                                                  \
    if (node == node->parent->child[0]) node->parent->child[0] = NULL;  \
    if (node == node->parent->child[1]) node->parent->child[1] = NULL;  \
    temp = node->parent; free(node); node = temp; continue;             \
}

// Deletes memory allocated for a binary tree object
static void lrzdFreeTree(LRZ_NODE *root) {
    LRZ_NODE *temp, *node = root;

    // Exit now if the tree wasn't initialized
    if (root == NULL) return;

    // Traverse the tree until all child nodes are deleted
    while (1) {

        // Terminal nodes are deleted outright
        if (node->type != NODE_BRANCHING) DeleteNode();

        // Delete a branching node if both its children are gone
        if (node->child[0] == NULL && node->child[1] == NULL) {
            if (node == root) break; // Root has no parent, so exit here
            DeleteNode();
        }

        // Select the next non-deleted child node
        if (node->child[0] != NULL)
            node = node->child[0];
        else node = node->child[1];
    }

    // Delete root and return
    free(root);
    return;
}

// Deletes memory allocated for a chunk object
static void lrzdFreeChunk(LRZ_CHUNK *chunk) {
    lrzdFreeTree(chunk->rawtree);
    lrzdFreeTree(chunk->lentree);
    lrzdFreeTree(chunk->disttree);
    return;
}

// Decodes a binary tree from a chunk header
static int lrzdGetTree(LRZ_CHUNK *chunk, LRZ_NODE **out) {
    int nibnum, x, nodenum = 3, donenum = 0, level = 1;
    unsigned char nibs[17], curbyte;
    LRZ_NODE *root, *node;

    // Error checking
    if (chunk->srcoff >= chunk->srclen) return 1;

    // Get the number of nibbles for this tree
    curbyte = chunk->src[chunk->srcoff++];
    nibnum = (curbyte & 0x0F) + 2;
    if (nibnum == 2) return 0; // No nibbles for this tree
    if (chunk->srcoff + (nibnum >> 1) >= chunk->srclen)
        return 1; // Not enough bytes for this tree's definition

    // Process all of the nibbles for this tree
    for (x = 0; x < nibnum; x++) {
        if (x & 1) curbyte = chunk->src[chunk->srcoff++];
        else       curbyte >>= 4;
        nibs[x] = curbyte & 0x0F;
    }

    // Initialize the tree
    root = *out           = calloc(sizeof(LRZ_NODE), 1);
    node = root->child[0] = calloc(sizeof(LRZ_NODE), 1);
    root->child[1]        = calloc(sizeof(LRZ_NODE), 1);
    root->type = NODE_BRANCHING;
    root->child[0]->parent = root;
    root->child[1]->parent = root;

    // Construct the remainder of the tree
    while (!root->done) {

        // Check if this node needs further processing
        if (node->done) {
            node = node->parent;
            level--;
            continue;
        }

        // If a branching node, step down into the tree a level
        if (node->type == NODE_BRANCHING) {

            // Check if both child nodes are finished
            if (node->child[0]->done && node->child[1]->done) {
                node->done = 1;
                donenum++;
                continue;
            }

            // Step into the next unfinished child node
            if (node->child[0]->type == NODE_UNKNOWN)
                node = node->child[0];
            else node = node->child[1];
            level++;
            continue;
        }

        // If all the nibbles are less than the current level, return error
        for (x = 0; x < nibnum; x++)
            if (nibs[x] >= level) break;
        if (x == nibnum) return 1;

        // Try to find a value for this node
        for (x = 0; x < nibnum; x++)
            if (nibs[x] == level) break;

        // A value was found for this node
        if (x < nibnum) {
            nibs[x] = 0;
            node->type = NODE_VALUE;
            node->value = x;
            node->done = 1;
            donenum++;
            continue;
        }

        // No value could be found for this node; make it a branching node
        node->type = NODE_BRANCHING;
        node->child[0] = calloc(sizeof(LRZ_NODE), 1);
        node->child[1] = calloc(sizeof(LRZ_NODE), 1);
        node->child[0]->parent = node;
        node->child[1]->parent = node;
        nodenum += 2;
    }

    // If not all the nodes created were finalized, return error
    if (donenum != nodenum) return 1;

    // Return success
    return 0;
}

// Returns the value of the specified number of bits from the input stream
static int lrzdGetBits(LRZ_CHUNK *chunk, int count) {
    int ret;

    // Error checking
    if (chunk->bitsleft < count || count > 15) return -1;
    chunk->bitsleft -= count;

    // Load the bits from the stream into the register
    while (count > chunk->subbits) {
        chunk->reg <<= chunk->subbits;
        if (chunk->srcoff < chunk->srclen)
            chunk->reg |= (int) chunk->src[chunk->srcoff++];
        count -= chunk->subbits;
        chunk->subbits = 8;
    }
    chunk->reg <<= count;
    chunk->subbits -= count;

    // Process the bits from the register
    ret = (chunk->reg & 0x00FFFF00) >> 8;
    chunk->reg &= 0x000000FF;
    return ret;
}

// Loads a number from input data encoded as a normal value
static int lrzdParse(LRZ_CHUNK *chunk) {
    int x = 1, count = 0;

    // Read (and count) bits until a 0 is encountered
    while (x) {
        x = lrzdGetBits(chunk, 1);
        if (x < 0) return -1; // Error reading bit
        count += x;
    }

    // Read that many additional bits from the input
    x = lrzdGetBits(chunk, count);
    if (x < 0) return -1; // Error reading bits

    // Return the composited value
    return (1 << count) - 1 + x;
}

// Loads a number from input data by traversing a binary tree
static int lrzdTraverse(LRZ_CHUNK *chunk, LRZ_NODE *node) {
    int x, ret;

    // Traverse the tree to a terminal node
    while (node->type == NODE_BRANCHING) {
        x = lrzdGetBits(chunk, 1);
        if (x < 0) return -1; // Error reading bit
        node = node->child[x];
    }

    // Return the number as a literal
    x = node->value;
    if (x < 2) return x;

    // Read additional bits from the data
    ret = 1 << --x;
    x = lrzdGetBits(chunk, x);
    if (x < 0) return -1; // Error reading bits
    return ret | x;
}

// This macro is used by lrzdChunk() to load a number from the input stream
#define GetNumber(var, tree) {           \
    if (tree != NULL)                    \
        var = lrzdTraverse(chunk, tree); \
    else var = lrzdParse(chunk);         \
}

// Processes a file chunk
static int lrzdChunk(LRZ_CHUNK *chunk) {
    int x, dist, len;

    // Initialize input/output variables
    chunk->srcoff = 0;
    chunk->dstoff = 0;

    // Process binary trees
    chunk->rawtree = chunk->lentree = chunk->disttree = NULL;
    if (lrzdGetTree(chunk, &chunk->rawtree))  return 1;
    if (lrzdGetTree(chunk, &chunk->lentree))  return 1;
    if (lrzdGetTree(chunk, &chunk->disttree)) return 1;

    // Initialize bit processing variables
    chunk->bitsleft = (chunk->srclen - chunk->srcoff) * 8;
    if (!chunk->bitsleft) return 1; // Insufficient data
    chunk->reg = (int) chunk->src[chunk->srcoff++];
    chunk->subbits = 8;

    // Process input data until all output data is written
    chunk->state = STATE_RAW;
    while (chunk->dstoff < chunk->dstlen) {

        // Read raw bytes from the input
        if (chunk->state == STATE_RAW) {

            // Read the number of bytes to copy
            GetNumber(x, chunk->rawtree);
            if (x < 0 || chunk->dstoff + x > chunk->dstlen)
                return LRZ_BADDATA;

            // Copy that many bytes
            for (len = x; len > 0; len--) {
                x = lrzdGetBits(chunk, 8);
                if (x < 0) return LRZ_BADDATA;
                chunk->dst[chunk->dstoff++] = 
                    (unsigned char) x;
            }
        }

        // Copy data with a distance/length pair
        else {

            // Read the length and distance values
            GetNumber(len,  chunk->lentree);
            GetNumber(dist, chunk->disttree);
            if (len < 0 || dist < 0) return LRZ_BADDATA;
            len += 2; dist++; // Implicit processing
            if (chunk->dstoff + len > chunk->dstlen || 
                chunk->dstoff - dist < 0) 
                return LRZ_BADDATA;

            // Copy the bytes from earlier in the output
            for (x = chunk->dstoff - dist; len > 0; len--)
                chunk->dst[chunk->dstoff++] = chunk->dst[x++];
        }

        // Alternate processing states
        chunk->state = 1 - chunk->state;
    }

    // Return success
    return 0;
}



////////////////////////////////////////////////////////////////////////////////
//                               API Functions                                //
////////////////////////////////////////////////////////////////////////////////

// Decompresses a compressed buffer
int lrzDecode(unsigned char *src, int srclen, unsigned char *dst, int dstlen) {
    int chunks, packed, unpacked, srcoff = 1, dstoff = 0, x;
    LRZ_CHUNK chunk;

    // Error checking
    if (src == NULL || dst == NULL || srclen < 1 || dstlen < 1)
        return LRZ_BADPARAMS;

    // Read the chunk count and check input size
    chunks = (int) src[0];
    if (!chunks || chunks * 5 + 1 > srclen) return LRZ_BADDATA;

    // Process all chunks
    for (; chunks > 0; chunks--, srcoff += packed, dstoff += unpacked) {

        // Process the chunk header
        packed   = ((int) src[srcoff + 1] << 8) | (int) src[srcoff];
        if (srcoff + packed > srclen) return LRZ_BADDATA;
        unpacked = ((int) src[srcoff + 3] << 8) | (int) src[srcoff + 2];
        if (dstoff + unpacked > dstlen) return LRZ_BADDATA;

        // Construct and process a chunk object
        chunk.src    = &src[srcoff + 5]; chunk.dst    = &dst[dstoff];
        chunk.srclen = packed - 5;       chunk.dstlen = unpacked;
        x = lrzdChunk(&chunk); lrzdFreeChunk(&chunk);
        if (x) return LRZ_BADDATA;
    }

    // Return success
    return LRZ_SUCCESS;
}

// Examines a compressed buffer to determine its decompressed size
int lrzGetSize(unsigned char *data, int len) {
    int chunks, packed, unpacked = 0, off = 1;

    // Error checking
    if (data == NULL || len < 1) return LRZ_BADPARAMS;

    // Read the chunk count and check input size
    chunks = (int) data[0];
    if (!chunks || chunks * 5 + 1 > len) return LRZ_BADDATA;

    // Check each chunk's packed size
    for (; chunks > 0; chunks--, off += packed) {
        packed = ((int) data[off + 1] << 8) | (int) data[off];
        if (off + packed > len) return LRZ_BADDATA;
        unpacked += (((int) data[off + 3] << 8) | (int) data[off + 2]);
    }

    // Return the accumulated unpacked size
    return unpacked;
}
